



if not exist %PREFIX%\share\doc\heavyai\LICENSE.md exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\ThirdParty\gdal-data exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\completion_hints.thrift exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\heavy.thrift exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\common.thrift exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine/serialized_result_set.thrift exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine/extension_functions.thrift exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\Shared\funcannotations.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\Shared\InlineNullValues.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\Logger\Logger.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine\heavydbTypes.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine\RuntimeFunctions.bc exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine\GeosRuntime.bc exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\QueryEngine\ExtensionFunctions.ast exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\bin\heavyai-utility-6.0.0.jar exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\bin\heavyai-jdbc-6.0.0.jar exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\bin\calcite-1.0-SNAPSHOT-jar-with-dependencies.jar exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\bin\generate_cert exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
